"""Provides a widget for creating and displaying a Plotext plot."""

from __future__ import annotations

from rich.text import Text
from textual.app import RenderResult
from textual.reactive import var
from textual.widget import Widget

from .plot import Plot, ThemeName


class PlotextPlot(Widget):
    """A Plotext plot display widget."""

    DEFAULT_CSS = """
    PlotextPlot {
        width: 1fr;
        height: 1fr;
    }
    """

    auto_theme: var[bool] = var(True)
    """Should the plot automatically adjust the theme depending on light and dark mode?

    If set to `True` the theme of the plot will be adjusted to either
    [`light_mode_theme`][PlotextPlot.light_mode_theme] or
    [`dark_mode_theme`][PlotextPlot.dark_mode_theme] when
    [`App.dark`](https://textual.textualize.io/api/app/#textual.app.App.dark)
    is changed.

    By default this is set to `False` and `light_mode_theme` and
    `dark_mode_theme` are both set to the same theme, designed to take on
    colours appropriate to the current mode.
    """

    light_mode_theme: var[ThemeName] = var("textual-design-light")
    """The Plotext theme to use for light mode.

    Note:
        This theme is only used when [`auto_theme`][PlotextPlot.auto_theme] is `True`.
    """

    dark_mode_theme: var[ThemeName] = var("textual-design-dark")
    """The Plotext theme to use for dark mode.

    Note:
        This theme is only used when [`auto_theme`][PlotextPlot.auto_theme] is `True`.
    """

    def __init__(
        self,
        *,
        name: str | None = None,
        id: str | None = None,  # pylint:disable=redefined-builtin
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialise the Plotext plot widget.

        Args:
            name: The name of the Plotext plot widget.
            id: The ID of the Plotext plot widget in the DOM.
            classes: The CSS classes of the Plotext plot widget.
            disabled: Whether the Plotext plot widget is disabled or not.
        """
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self._plot = Plot()
        # We use textual-default as the default theme, as that's going to
        # work well no matter if we're in light or dark mode.
        self._plot.theme("textual-default")
        # Watch the application's dark mode switch so that we can react to
        # any request to auto-change between light and dark themes.
        self.watch(self.app, "dark", self._dark_mode, init=False)

    @property
    def plt(self) -> Plot:
        """The Plotext plotting object.

        Whereas normally Plotext-using code would so something like this:

        ```python
        import plotext as plt

        ...

        plt.some_call()
        plt.some_other_call()
        ```

        use this `plt` property instead of importing `plotext`.
        """
        return self._plot

    def render(self) -> RenderResult:
        """Render the plot.

        Returns:
            The renderable for displaying the plot.
        """
        self._plot.plotsize(self.size.width, self.size.height)
        # This is a belt-and-braces setting of the size of the plot.
        # Internally plotsize calls _set_plot, and as best as I can figure
        # out, what I'm doing here *should* be a no-op (or rather a repeat
        # of what I've just done). And yet... the resizing of plots just
        # doesn't work right without this. This *might* be some
        # as-yet-undiscovered side-effect of dumpster-diving for the figure
        # class.
        #
        # https://github.com/Textualize/textual-plotext/issues/5
        self._plot._set_size(self.size.width, self.size.height)
        if self.auto_theme:
            self._plot.theme(
                self.dark_mode_theme if self.app.dark else self.light_mode_theme
            )
        return Text.from_ansi(self._plot.build())

    def _watch_light_mode_theme(self) -> None:
        """React to changes to the light mode theme."""
        if self.auto_theme and not self.app.dark:
            self.refresh()

    def _watch_dark_mode_theme(self) -> None:
        """React to changes to the dark mode theme."""
        if self.auto_theme and self.app.dark:
            self.refresh()

    def _dark_mode(self) -> None:
        """React to dark mode being toggled."""
        if self.auto_theme:
            self.refresh()



"""Experimental type stub for Plotext.

To start with this simply contains what's needed to help fully type-check
the code of this library; but as time goes on this could turn into a set of
stubs for the whole of Plotext.
"""

from __future__ import annotations

from typing import Any

def sin(
    periods: int,
    length: int,
    amplitude: int,
    phase: float,
    decay: int,
) -> list[float]: ...
def square(periods: int, length: int, amplitude: float) -> list[int]: ...
def colorize(
    string: str,
    fullground: str | int | tuple[int, int, int] | None,
    style: str | None,
    background: str | int | tuple[int, int, int] | None,
    show: bool,
) -> str: ...
def uncolorize(string: str) -> str: ...
def transpose(data: list[list[Any]]) -> list[list[Any]]: ...

platform: str = ...



"""Type hints for the methods of the Plotext _figure_class class.

NOTE: Some of the work in here is guesswork. There are no type hints inside
Plotext's code (hence this stub file), and it's not always clear from the
Plotext code what the type of any given parameter on any given method is.

Corrections are most welcome!

ALSO NOTE: For the moment this is simply for support while developing
textual-plotext; but this could turn into something that provides more
general support for the type-hinting of Plotext.
"""

from __future__ import annotations

from datetime import datetime
from typing import Any, Sequence, Tuple, Union

from typing_extensions import Literal, Self, TypeAlias

Alignment: TypeAlias = Literal["left", "center", "right", "top", "bottom", "dynamic"]
Color: TypeAlias = Union[str, int, Tuple[int, int, int]]
Orientation: TypeAlias = Literal["horizontal", "vertical"]

class _figure_class:
    ############################################################################
    # Subplots Functions
    def subplot(
        self, row: int | None = None, col: int | None = None
    ) -> _figure_class: ...
    def subplots(self, rows: int | None = None, cols: int | None = None) -> Self: ...

    ############################################################################
    # External Set Functions
    def title(self, title: str | None = None) -> None: ...
    def xlabel(self, label: str | None = None, xside: str | None = None) -> None: ...
    def ylabel(self, label: str | None = None, xside: str | None = None) -> None: ...
    def xlim(
        self,
        left: float | None = None,
        right: float | None = None,
        xside: float | None = None,
    ) -> None: ...
    def ylim(
        self,
        lower: float | None = None,
        upper: float | None = None,
        yside: float | None = None,
    ) -> None: ...
    def xscale(self, scale: str | None = None, xside: str | None = None) -> None: ...
    def yscale(self, scale: str | None = None, yside: str | None = None) -> None: ...
    def xticks(
        self,
        ticks: list[float] | None = None,
        labels: list[str] | None = None,
        xside: float | None = None,
    ) -> None: ...
    def yticks(
        self,
        ticks: list[float] | None = None,
        labels: list[str] | None = None,
        yside: float | None = None,
    ) -> None: ...
    def xfrequency(
        self, frequency: float | None = None, xside: float | None = None
    ) -> None: ...
    def yfrequency(
        self, frequency: float | None = None, yside: float | None = None
    ) -> None: ...
    def xreverse(
        self, reverse: bool | None = None, xside: float | None = None
    ) -> None: ...
    def yreverse(
        self, reverse: bool | None = None, yside: float | None = None
    ) -> None: ...
    def xaxes(
        self, lower: bool | int | None = None, upper: bool | int | None = None
    ) -> None: ...
    def yaxes(
        self, left: bool | int | None = None, right: bool | int | None = None
    ) -> None: ...
    def frame(self, frame: bool | int | None = None) -> None: ...
    def grid(
        self, horizontal: bool | int | None = None, vertical: bool | int | None = None
    ) -> None: ...
    def canvas_color(self, color: Color | None = None) -> None: ...
    def axes_color(self, color: Color | None = None) -> None: ...
    def ticks_color(self, color: Color | None = None) -> None: ...
    def ticks_style(self, style: Color | None = None) -> None: ...
    def theme(self, theme: str | None = None) -> None: ...

    # Clear Functions
    def clear_figure(self) -> None: ...
    def clf(self) -> None: ...
    def clear_data(self) -> None: ...
    def cld(self) -> None: ...
    def clear_color(self) -> None: ...
    def clc(self) -> None: ...
    def clear_terminal(self, lines: int | None = None) -> None: ...
    def clr(self, lines: int | None = None) -> None: ...

    ############################################################################
    # Plot Functions
    def scatter(
        self,
        *args: Sequence[Any],
        marker: str | None = None,
        color: Color | None = None,
        style: str | None = None,
        fillx: float | bool | str | None = None,
        filly: float | bool | str | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...
    def plot(
        self,
        *args: Sequence[Any],
        marker: str | None = None,
        color: Color | None = None,
        style: str | None = None,
        fillx: float | bool | str | None = None,
        filly: float | bool | str | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...
    def bar(
        self,
        *args: Sequence[Any],
        xside: str | None = None,
        yside: str | None = None,
        marker: str | None = None,
        color: Color | None = None,
        fill: bool | None = None,
        width: float | None = None,
        orientation: Orientation | None = None,
        label: str | None = None,
        minimum: float | None = None,
        reset_ticks: bool | None = None,
    ) -> None: ...
    def multiple_bar(
        self,
        *args: Sequence[Any],
        xside: str | None = None,
        yside: str | None = None,
        marker: str | None = None,
        color: Color | None = None,
        fill: bool | None = None,
        width: float | None = None,
        orientation: Orientation | None = None,
        label: str | None = None,
        minimum: float | None = None,
        reset_ticks: bool | None = None,
    ) -> None: ...
    def stacked_bar(
        self,
        *args: Sequence[Any],
        xside: str | None = None,
        yside: str | None = None,
        marker: str | None = None,
        color: Color | None = None,
        fill: bool | None = None,
        width: float | None = None,
        orientation: Orientation | None = None,
        label: str | None = None,
        minimum: float | None = None,
        reset_ticks: bool | None = None,
    ) -> None: ...
    def hist(
        self,
        data: Sequence[Any],
        bins: int | None = None,
        marker: str | None = None,
        color: Color | None = None,
        fill: bool | None = None,
        norm: bool = False,
        width: float | None = None,
        orientation: Orientation | None = None,
        minimum: float | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...
    def candlestick(
        self,
        dates: Sequence[str],
        data: dict[str, Any],
        colors: list[Color] | None = None,
        orientation: Orientation | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...

    ############################################################################
    # Plotting Tools
    def error(
        self,
        *args: Sequence[Any],
        xerr: Sequence[float] | None = None,
        yerr: Sequence[float] | None = None,
        color: Color | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...
    def event_plot(
        self,
        data: Sequence[str],
        marker: str | None = None,
        color: Color | None = None,
        orientation: Orientation | None = None,
        side: str | None = None,
    ) -> None: ...
    def eventplot(
        self,
        data: Sequence[str],
        marker: str | None = None,
        color: Color | None = None,
        orientation: Orientation | None = None,
        side: str | None = None,
    ) -> None: ...
    def vertical_line(
        self, coordinate: float, color: Color | None = None, xside: str | None = None
    ) -> None: ...
    def vline(
        self, coordinate: float, color: Color | None = None, xside: str | None = None
    ) -> None: ...
    def horizontal_line(
        self, coordinate: float, color: Color | None = None, yside: str | None = None
    ) -> None: ...
    def hline(
        self, coordinate: float, color: Color | None = None, yside: str | None = None
    ) -> None: ...
    def text(
        self,
        text: str,
        x: float | None = None,
        y: float | None = None,
        color: Color | None = None,
        background: Color | None = None,
        style: str | None = None,
        orientation: Orientation | None = None,
        alignment: Alignment | None = None,
        xside: str | None = None,
        yside: str | None = None,
    ) -> None: ...
    def rectangle(
        self,
        x: float | None = None,
        y: float | None = None,
        marker: str | None = None,
        color: Color | None = None,
        lines: bool | None = None,
        fill: bool | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...
    def polygon(
        self,
        x: float | None = None,
        y: float | None = None,
        radius: float | None = None,
        sides: int | None = None,
        marker: str | None = None,
        color: Color | None = None,
        lines: bool | None = None,
        fill: bool | None = None,
        xside: str | None = None,
        yside: str | None = None,
        label: str | None = None,
    ) -> None: ...
    def confusion_matrix(
        self,
        actual: Sequence[float],
        predicted: Sequence[float],
        color: Color | None = None,
        style: str | None = None,
        labels: Sequence[str] | None = None,
    ) -> None: ...
    def cmatrix(
        self,
        actual: Sequence[float],
        predicted: Sequence[float],
        color: Color | None = None,
        style: str | None = None,
        labels: Sequence[str] | None = None,
    ) -> None: ...
    def indicator(
        self,
        value: float | str,
        label: str | None = None,
        color: Color | None = None,
        style: str | None = None,
    ) -> None: ...

    ############################################################################
    # 2D Plots
    def matrix_plot(
        self,
        matrix: Sequence[Sequence[float | tuple[int, int, int]]],
        marker: str | None = None,
        style: str | None = None,
        fast: bool = False,
    ) -> None: ...
    def image_plot(
        self,
        path: str,
        marker: str | None = None,
        style: str | None = None,
        fast: bool = False,
        grayscale: bool = False,
    ) -> None: ...

    ############################################################################
    # Date Functions
    def date_form(
        self, input_form: str | None = None, output_form: str | None = None
    ) -> None: ...
    def set_time0(self, string: str, input_form: str | None = None) -> None: ...
    def today_datetime(self) -> datetime: ...
    def today_string(self, output_form: str | None = None) -> str: ...
    def datetime_to_string(
        self, datetime: datetime, output_form: str | None = None
    ) -> str: ...
    def datetimes_to_string(
        self, datetimes: Sequence[datetime], output_form: str | None = None
    ) -> list[str]: ...
    def string_to_datetime(
        self, string: str, input_form: str | None = None
    ) -> datetime: ...
    def string_to_time(
        self, string: str, input_form: str | None = None
    ) -> datetime: ...
    def strings_to_time(
        self, string: Sequence[str], input_form: str | None = None
    ) -> datetime: ...

    ############################################################################
    # Build Functions
    def build(self) -> str: ...

    ############################################################################
    # Externally Called Utility
    def main(self) -> _figure_class: ...
    def plot_size(
        self, width: int | None = None, height: int | None = None
    ) -> None: ...
    def plotsize(self, width: int | None = None, height: int | None = None) -> None: ...
    def _set_size(
        self, width: int | None = None, height: int | None = None
    ) -> None: ...
    def take_min(self) -> None: ...
